import 'dart:async';
import 'dart:io';
import 'dart:typed_data';

import 'package:http/http.dart' as http;
import 'package:yona/src/addr.dart';
import 'package:yona_crypto/b64.dart';
import 'package:yona_crypto/crypto.dart';

class Inbox {
  final RawDatagramSocket _socket;
  final int _port;
  final InternetAddress _host;
  late final Timer _timer;

  Inbox(this._port, this._host, this._socket) {
    _timer = Timer.periodic(Duration(milliseconds: 100), (timer) {
      _check(_last);
    });
  }

  static Future<Inbox> connect(NetAddr addr) async {
    var socket = await RawDatagramSocket.bind(InternetAddress.anyIPv4, 0);
    var resp =
        await InternetAddress.lookup(addr.host, type: socket.address.type);
    return Inbox(addr.port, resp.first, socket);
  }

  void _check(List<int> last) {
    _socket.send(last, _host, _port);
  }

  Uint8List _last = Uint8List(1);

  Stream<Packet> get stream {
    return _socket
        .map((_) => _socket.receive())
        .where((dg) => dg != null)
        .map((dg) {
      _last = dg!.data;
      return Packet(dg.data);
    });
  }

  Future<void> put(Packet packet, B64 sk) async {
    var sig = sign(sk, packet.bytes);
    var resp = await http.put(
        Uri(scheme: "http", host: _host.address, port: _port),
        body: packet.bytes,
        headers: {"Content-Signature": sig.toString()});
    if (resp.statusCode != 204) {
      throw Exception(resp.statusCode);
    }
  }

  void close() {
    if (_timer.isActive) _timer.cancel();
    return _socket.close();
  }
}

class Packet {
  Uint8List bytes;

  Packet(this.bytes);

  static Packet create(List<int> from, List<int> to, List<int> data, int id) {
    if (from.length != 32 || to.length != 32) {
      throw ArgumentError("");
    }
    var n = Uint8List(4);
    n.buffer.asByteData().setUint32(0, id, Endian.little);
    return Packet(Uint8List.fromList(from + to + n.toList() + data));
  }

  Uint8List get from {
    return bytes.sublist(0, 32);
  }

  Uint8List get to {
    return bytes.sublist(32, 64);
  }

  Uint8List get id {
    return bytes.sublist(64, 68);
  }

  Uint8List get data {
    return bytes.sublist(68);
  }
}

void main() async {
  var addr = NetAddr.fromString("localhost:6666");
  var box = await Inbox.connect(addr);
  box.stream.map((event) => event.id).forEach(print);
  var sk = genSK();
  var pk = derivePK(sk).decode().sublist(0, 32);
  Timer.periodic(Duration(milliseconds: 100), (timer) async {
    await box.put(Packet.create(pk, pk, [0], timer.tick), sk);
  });
}
