import 'dart:async';
import 'dart:convert';

import 'package:http/http.dart' as http;
import 'package:yona/src/addr.dart';
import 'package:yona/src/webinbox.dart';
import 'package:yona/src/webroker.dart';
import 'package:yona_crypto/b64.dart';
import 'package:yona_crypto/crypto.dart';

import 'msg.dart';
import 'webinbox.dart';

abstract class Handlers {
  Future<void> onError<T>(T error, [StackTrace? trace]);

  Future<void> onMsg(Msg msg);
}

class Node extends Peer {
  final B64 sk;
  final Handlers _handlers;

  Node(
    this._handlers, {
    required this.sk,
    required this.webrokers,
    required String name,
    required NetAddr verival,
    required NetAddr webinbox,
  }) : super(derivePK(sk), verival, name: name, webinbox: webinbox);

  Future<void> publish() async {
    var payload = toString();
    var sig = sign(sk, utf8.encoder.convert(payload));
    var resp = await http.put(
        Uri(scheme: "http", host: verival.host, port: verival.port),
        body: "$sig&$payload");
    if (resp.statusCode != 204) {
      throw Exception(resp.body.isEmpty ? resp.statusCode : resp.body);
    }
  }

  final List<NetAddr> webrokers;

  NetAddr _nextWebroker() {
    var addr = webrokers.first;
    webrokers.removeAt(0);
    webrokers.add(addr);
    return addr;
  }

  @override
  Target? get webroker {
    return _webrokerClient?.target;
  }

  Webroker? _webrokerClient;

  Future<void> _input(String msg) async {
    try {
      return await _inputMsg(IOMsg.fromString(msg));
    } catch (error) {
      await _handlers.onError(error);
    }
  }

  Future<void> _inputMsg(IOMsg input) async {
    try {
      var msg = await input.decrypt(sk);
      await _handlers.onMsg(msg);
    } catch (error) {
      await _handlers.onError(error);
    }
  }

  Future<void> send(BigInt id, String data, Peer peer, bool persist) async {
    await peer.fetch();
    var output = await IOMsg.fromMsg(Msg(id, data, this), sk, peer.pk);
    if (peer.webroker != null) {
      Webroker.send(peer.webroker!, output.toString()).ignore();
    }
    if (persist) {
      await Webinbox.put(peer.webinbox!, output, sk);
    }
  }

  bool _closed = false;

  Future<void> listenWebinbox() async {
    while (!_closed) {
      var client = Webinbox(pk, webinbox!);
      try {
        await client.stream.forEach((msg) => _inputMsg(msg));
      } catch (error) {
        _handlers.onError(error);
      } finally {
        client.close();
      }
    }
  }

  bool get closed => _closed;

  Future<void> listenWebroker() async {
    while (!_closed) {
      var addr = _nextWebroker();
      try {
        _webrokerClient = await Webroker.connect(addr);
        await publish();
        await _webrokerClient!.stream.forEach((str) => _input(str));
      } catch (error) {
        _handlers.onError(error);
      } finally {
        _webrokerClient?.close();
      }
    }
  }

  void close() {
    _closed = true;
  }
}

class Peer {
  final B64 pk;
  final NetAddr verival;
  NetAddr? _webinbox;
  Target? _webroker;
  String? _name;

  Peer(this.pk, this.verival,
      {String? name, NetAddr? webinbox, Target? webroker}) {
    _name = name;
    _webinbox = webinbox;
    _webroker = webroker;
  }

  Target? get webroker {
    return _webroker;
  }

  NetAddr? get webinbox {
    return _webinbox;
  }

  String get name {
    return _name ?? pk.toString();
  }

  static Peer fromMap(Map<String, String> data) {
    return Peer(
        //
        B64(data["pk"]!),
        //
        NetAddr.fromString(data["verival"]!),
        //
        name: data["name"] == null ? null : data["name"]!,
        //
        webinbox: data["webinbox"] == null
            ? null
            : NetAddr.fromString(data["webinbox"]!),
        //
        webroker: data["webroker"] == null
            ? null
            : Target.fromString(data["webroker"]!));
  }

  static Peer fromString(String peer) {
    var data = Uri.splitQueryString(peer);
    return Peer.fromMap(data);
  }

  Future<void> fetch() async {
    var res = await http.get(Uri(
        scheme: "http",
        host: verival.host,
        port: verival.port,
        pathSegments: [pk.toString()]));
    if (res.statusCode != 200) {
      throw Exception(res.body.isEmpty ? res.statusCode : res.body);
    }
    var ampersandIndex = res.body.indexOf('&');
    var signature = B64(res.body.substring(0, ampersandIndex));
    var payload = res.body.substring(ampersandIndex + 1);
    var data = Uri.splitQueryString(payload);
    verify(utf8.encoder.convert(payload), signature, B64(data["pk"]!));
    var peer = Peer.fromMap(data);
    _webinbox = peer.webinbox ?? webinbox;
    _webroker = peer.webroker ?? webroker;
    _name = peer._name ?? _name;
  }

  Map<String, String> toMap() {
    var map = {
      "pk": pk.toString(),
      "verival": verival.toString(),
      "time": (DateTime.now().microsecondsSinceEpoch ~/ 1000000).toString()
    };
    if (webroker != null) {
      map["webroker"] = webroker.toString();
    }
    if (_webinbox != null) {
      map["webinbox"] = webinbox.toString();
    }
    if (_name != null) {
      map["name"] = name;
    }
    return map;
  }

  @override
  String toString() {
    return Uri(queryParameters: toMap()).query;
  }
}
