import 'dart:async';
import 'dart:convert';

import 'package:http/http.dart' as http;
import 'package:yona/src/addr.dart';
import 'package:yona/src/msg.dart';
import 'package:yona_crypto/b64.dart';
import 'package:yona_crypto/crypto.dart';

class Webinbox {
  final NetAddr addr;
  final B64 pk;
  bool _closed = false;

  late BigInt _last;

  BigInt get last {
    return _last;
  }

  Webinbox(this.pk, this.addr, [BigInt? last]) {
    _last = last ?? BigInt.zero;
  }

  void close() {
    _closed = true;
  }

  bool _isPolling = false;

  Stream<IOMsg> get stream {
    return () async* {
      while (!_closed) {
        var msg = await _poll();
        if (msg == null) {
          await Future.delayed(Duration(seconds: 1));
        } else {
          yield msg;
        }
      }
    }();
  }

  Future<IOMsg?> _poll() async {
    if (_isPolling) return null;
    _isPolling = true;
    try {
      var resp = await http.get(Uri(
          scheme: "http",
          host: addr.host,
          port: addr.port,
          queryParameters: {"since": _last.toString()},
          pathSegments: [pk.toString()]));
      if (resp.statusCode != 200) {
        return null;
      }
      var msg = IOMsg.fromString(resp.body, pk);
      _last = msg.id ?? _last;
      return msg;
    } catch (_) {
      return null;
    } finally {
      _isPolling = false;
    }
  }

  static Future<void> put(NetAddr addr, IOMsg msg, B64 sk) async {
    var sig = sign(sk, utf8.encoder.convert(msg.data));
    var resp = await http.put(
        Uri(scheme: "http", host: addr.host, port: addr.port),
        headers: {"Content-Signature": sig.toString()},
        body: msg.toString());
    if (resp.statusCode != 204) {
      throw Exception(resp.body.isEmpty ? resp.statusCode : resp.body);
    }
  }
}

void main() async {
  var sk = genSK();
  var pk2 = derivePK(genSK());
  var addr = NetAddr.fromString("yonafediverse.com:4444");
  Timer.periodic(Duration(milliseconds: 500), (timer) async {
    await Webinbox.put(
        addr, IOMsg(derivePK(sk), pk2, DateTime.now().toString()), sk);
  });
  var cli = Webinbox(pk2, addr);
  Timer(Duration(seconds: 3), () => cli.close());
  await cli.stream.forEach(print);
  print("end");
}
