import 'dart:async';
import 'dart:convert';

import 'package:async/async.dart';
import 'package:yona/src/addr.dart';

import 'sse.dart' as sse;

class Webroker {
  final sse.SSE _sse;
  final String id;
  final NetAddr addr;
  Stream<String> _stream;
  num _lastEventTime;

  Target get target {
    return Target(addr, id);
  }

  void _updateLastEventTime() {
    _lastEventTime = DateTime.now().microsecondsSinceEpoch;
  }

  bool _isPinging = false;

  Future<void> _ping() async {
    try {
      if (_isPinging ||
          DateTime.now().microsecondsSinceEpoch - _lastEventTime < 1000000) {
        return;
      }
      _isPinging = true;
      await send(target, "ping");
    } finally {
      _isPinging = false;
    }
  }

  late Timer _timer;

  Webroker(this.id, this._stream, this._sse, this.addr, this._lastEventTime) {
    _stream = _stream.map((event) {
      _updateLastEventTime();
      return event;
    });
    _timer = Timer.periodic(Duration(seconds: 1), (timer) async {
      try {
        await _ping();
      } catch (err) {
        close();
      }
    });
  }

  static Future<Webroker> connect(NetAddr addr) async {
    var client = await sse.connect(addr);
    var q = StreamQueue<String>(client.stream
        .timeout(Duration(seconds: 3), onTimeout: (_) => client.close())
        .map((event) => base64.decoder.convert(event))
        .map((data) => utf8.decoder.convert(data))
        .where((str) => str != "ping"));
    return Webroker(await q.next, q.rest, client, addr, 0);
  }

  Stream<String> get stream {
    return _stream;
  }

  void close() {
    if (_timer.isActive) _timer.cancel();
    return _sse.close();
  }

  static Future<void> send(Target target, String data) async {
    var code = await sse.send(target.addr, target.id, data);
    if (code != 204) {
      throw Exception(code);
    }
  }
}

class Target {
  final NetAddr addr;
  final String id;

  Target(this.addr, this.id);

  @override
  String toString() {
    return "$addr/$id";
  }

  static Target fromString(String target) {
    var parts = target.split("/");
    if (parts.length != 2) {
      throw Exception("invalid target");
    }
    return Target(NetAddr.fromString(parts[0]), parts[1]);
  }
}

void main() async {
  var w = await Webroker.connect(NetAddr.fromString("yonafediverse.com:2222"));
  await w.stream.handleError(print).forEach((data) => print(":$data"));
  w.close();
}
